#//----------------------------------------------------------------------
#// Copyright 2011-2014 Mentor Graphics Corporation
#// Copyright 2014 Semifore
#// Copyright 2014 Intel Corporation
#// Copyright 2010-2017 Synopsys, Inc.
#// Copyright 2011-2018 Cadence Design Systems, Inc.
#// Copyright 2014-2018 NVIDIA Corporation
#// Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from ..macros.uvm_message_defines import (uvm_error, uvm_fatal)

#// CLASS -- NODOCS -- UVMTLMTime
#// Canonical time type that can be used in different timescales
#//
#// This time type is used to represent time values in a canonical
#// form that can bridge initiators and targets located in different
#// timescales and time precisions.
#//
#// For a detailed explanation of the purpose for this class,
#// see <Why is this necessary>.


class UVMTime:

    #   static local real m_resolution = 1.0e-12; // ps by default
    m_resolution = 1.0e-12
    #   local time m_time;  // Number of 'm_res' time units,
    #   local string m_name
    #
    #   // Function -- NODOCS -- set_time_resolution
    #   // Set the default canonical time resolution.
    #   //
    #   // Must be a power of 10.
    #   // When co-simulating with SystemC, it is recommended
    #   // that default canonical time resolution be set to the
    #   // SystemC time resolution.
    #   //
    #   // By default, the default resolution is 1.0e-12 (ps)
    #   //
    @classmethod
    def set_time_resolution(cls, res):
        # // Actually, it does not *really* need to be a power of 10.
        cls.m_resolution = res


    #   // Function -- NODOCS -- new
    #   // Create a new canonical time value.
    #   //
    #   // The new value is initialized to 0.
    #   // If a resolution is not specified,
    #   // the default resolution,
    #   // as specified by <set_time_resolution()>,
    #   // is used.
    def __init__(self, name="UVMTLMTime", res=0):
        self.m_name = name
        self.m_res = res
        self.m_time = 0
        if res == 0:
            self.m_res = UVMTime.m_resolution
        self.reset()



    #   // Function -- NODOCS -- get_name
    #   // Return the name of this instance
    #   //
    #   def string get_name(self):
    #      return m_name
    #   endfunction
    #
    #
    #   // Function -- NODOCS -- reset
    #   // Reset the value to 0
    def reset(self):
        self.m_time = 0

    #   // Scale a timescaled value to 'm_res' units,
    #   // the specified scale
    def to_m_res(self, t, scaled, secs):
        # ToDo: Check resolution
        return (t/scaled) * (secs/self.m_res)  # cast to 'real' removed
    #   endfunction


    #   // Function -- NODOCS -- get_realtime
    #   // Return the current canonical time value,
    #   // scaled for the caller's timescale
    #   //
    #   // ~scaled~ must be a time literal value that corresponds
    #   // to the number of seconds specified in ~secs~ (1ns by default).
    #   // It must be a time literal value that is greater or equal
    #   // to the current timescale.
    #   //
    #   //| #(delay.get_realtime(1ns));
    #   //| #(delay.get_realtime(1fs, 1.0e-15));
    #   //
    def get_realtime(self, scaled, secs=1.0e-9):
        return self.m_time * float(scaled) * self.m_res / secs

    #   // Function -- NODOCS -- incr
    #   // Increment the time value by the specified number of scaled time unit
    #   //
    #   // ~t~ is a time value expressed in the scale and precision
    #   // of the caller.
    #   // ~scaled~ must be a time literal value that corresponds
    #   // to the number of seconds specified in ~secs~ (1ns by default).
    #   // It must be a time literal value that is greater or equal
    #   // to the current timescale.
    #   //
    #   //| delay.incr(1.5ns, 1ns);
    #   //| delay.incr(1.5ns, 1ps, 1.0e-12);
    #   //
    def incr(self, t, scaled, secs=1.0e-9):
        if t < 0.0:
            uvm_error("UVM/TLM/TIMENEG", "Cannot increment UVMTLMTime variable " 
                    + self.m_name + " by a negative value")
            return

        if scaled == 0:
            uvm_fatal("UVM/TLM/BADSCALE",
                "UVMTLMTime::incr() called with a scaled time literal that is smaller than the current timescale")


        self.m_time += self.to_m_res(t, scaled, secs)

    #
    #   // Function -- NODOCS -- decr
    #   // Decrement the time value by the specified number of scaled time unit
    #   //
    #   // ~t~ is a time value expressed in the scale and precision
    #   // of the caller.
    #   // ~scaled~ must be a time literal value that corresponds
    #   // to the number of seconds specified in ~secs~ (1ns by default).
    #   // It must be a time literal value that is greater or equal
    #   // to the current timescale.
    #   //
    #   //| delay.decr(200ps, 1ns);
    #   //
    #   def void decr(self,real t, time scaled, real secs):
    #      if (t < 0.0):
    #         `uvm_error("UVM/TLM/TIMENEG", {"Cannot decrement UVMTLMTime variable ", m_name, " by a negative value"})
    #         return
    #      end
    #      if (scaled == 0):
    #         `uvm_fatal("UVM/TLM/BADSCALE",
    #                    "UVMTLMTime::decr() called with a scaled time literal that is smaller than the current timescale")
    #      end
    #
    #      m_time -= to_m_res(t, scaled, secs)
    #
    #      if (m_time < 0.0):
    #         `uvm_error("UVM/TLM/TOODECR", {"Cannot decrement UVMTLMTime variable ", m_name, " to a negative value"})
    #         reset()
    #      end
    #   endfunction
    #
    #
    #   // Function -- NODOCS -- get_abstime
    #   // Return the current canonical time value,
    #   // in the number of specified time unit, regardless of the
    #   // current timescale of the caller.
    #   //
    #   // ~secs~ is the number of seconds in the desired time unit
    #   // e.g. 1e-9 for nanoseconds.
    #   //
    #   //| $write("%.3f ps\n", delay.get_abstime(1e-12));
    #   //
    #   def real get_abstime(self,real secs):
    #      return m_time*m_res/secs
    #   endfunction
    #
    #
    #   // Function -- NODOCS -- set_abstime
    #   // Set the current canonical time value,
    #   // to the number of specified time unit, regardless of the
    #   // current timescale of the caller.
    #   //
    #   // ~secs~ is the number of seconds in the time unit in the value ~t~
    #   // e.g. 1e-9 for nanoseconds.
    #   //
    #   //| delay.set_abstime(1.5, 1e-12));
    #   //
    #   def void set_abstime(self,real t, real secs):
    #      m_time = t*secs/m_res
    #   endfunction
    #endclass

UVMTLMTime = UVMTime

#// Group -- NODOCS -- Why is this necessary
#//
#// Integers are not sufficient, on their own,
#// to represent time without any ambiguity:
#// you need to know the scale of that integer value.
#// That scale is information conveyed outside of that integer.
#// In SystemVerilog, it is based on the timescale
#// that was active when the code was compiled.
#// SystemVerilog properly scales time literals, but not integer values.
#// That's because it does not know the difference between an integer
#// that carries an integer value and an integer that carries a time value.
#// The 'time' variables are simply 64-bit integers,
#// they are not scaled back and forth to the underlying precision.
#//
#//| `timescale 1ns/1ps
#//|
#//| module m();
#//|
#//| time t;
#//|
#//| initial
#//| begin
#//|    #1.5;
#//|    $write("T=%f ns (1.5)\n", $realtime());
#//|    t = 1.5;
#//|    #t;
#//|    $write("T=%f ns (3.0)\n", $realtime());
#//|    #10ps;
#//|    $write("T=%f ns (3.010)\n", $realtime());
#//|    t = 10ps;
#//|    #t;
#//|    $write("T=%f ns (3.020)\n", $realtime());
#//| end
#//| endmodule
#//
#// yields
#//
#//| T=1.500000 ns (1.5)
#//| T=3.500000 ns (3.0)
#//| T=3.510000 ns (3.010)
#//| T=3.510000 ns (3.020)
#//
#// Within SystemVerilog, we have to worry about
#// - different time scale
#// - different time precision
#//
#// Because each endpoint in a socket
#// could be coded in different packages
#// and thus be executing under different timescale directives,
#// a simple integer cannot be used to exchange time information
#// across a socket.
#//
#// For example
#//
#//| `timescale 1ns/1ps
#//|
#//| package a_pkg;
#//|
#//| class a;
#//|    function void f(inout time t);
#//|       t += 10ns;
#//|    endfunction
#//| endclass
#//|
#//| endpackage
#//|
#//|
#//| `timescale 1ps/1ps
#//|
#//| program p;
#//|
#//| import a_pkg::*;
#//|
#//| time t;
#//|
#//| initial
#//| begin
#//|    a A = new;
#//|    A.f(t);
#//|    #t;
#//|    $write("T=%0d ps (10,000)\n", $realtime());
#//| end
#//| endprogram
#//
#// yields
#//
#//| T=10 ps (10,000)
#//
#// Scaling is needed every time you make a procedural call
#// to code that may interpret a time value in a different timescale.
#//
#// Using the UVMTLMTime type
#//
#//| `timescale 1ns/1ps
#//|
#//|    package a_pkg;
#//|
#//| import uvm_pkg::*;
#//|
#//| class a;
#//|    function void f(UVMTLMTime t);
#//|       t.incr(10ns, 1ns);
#//|    endfunction
#//| endclass
#//|
#//| endpackage
#//|
#//|
#//| `timescale 1ps/1ps
#//|
#//| program p;
#//|
#//| import uvm_pkg::*;
#//| import a_pkg::*;
#//|
#//| UVMTLMTime t = new;
#//|
#//| initial
#//|    begin
#//|       a A = new;
#//|       A.f(t);
#//|       #(t.get_realtime(1ns));
#//|       $write("T=%0d ps (10,000)\n", $realtime());
#//| end
#//| endprogram
#//
#// yields
#//
#//| T=10000 ps (10,000)
#//
#// A similar procedure is required when crossing any simulator
#// or language boundary,
#// such as interfacing between SystemVerilog and SystemC.



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc. 
#//   Copyright 2010 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------
#
#//------------------------------------------------------------------------------
#// CLASS: uvm_random_stimulus #(T)
#//
#// A general purpose unidirectional random stimulus class.
#//
#// The uvm_random_stimulus class generates streams of T transactions. These streams
#// may be generated by the randomize method of T, or the randomize method of
#// one of its subclasses.  The stream may go indefinitely, until terminated
#// by a call to stop_stimulus_generation, or we may specify the maximum number
#// of transactions to be generated.
#//
#// By using inheritance, we can add directed initialization or tidy up after
#// random stimulus generation. Simply extend the class and define the run task,
#// calling super.run() when you want to begin the random stimulus phase of
#// simulation.
#//
#// While very useful in its own right, this component can also be used as a
#// template for defining other stimulus generators, or it can be extended to
#// add additional stimulus generation methods and to simplify test writing.
#//
#//------------------------------------------------------------------------------
#
#class uvm_random_stimulus #(type T=uvm_transaction) extends uvm_component;
#
#  const static string type_name = "uvm_random_stimulus #(T)";
#
#  typedef uvm_random_stimulus #(T) this_type;
#  `uvm_component_param_utils(this_type)
#
#  // Port: blocking_put_port
#  //
#  // The blocking_put_port is used to send the generated stimulus to the rest
#  // of the testbench.
#                                                      
#  uvm_blocking_put_port #(T) blocking_put_port;
#
#
#  // Function: new
#  //
#  // Creates a new instance of a specialization of this class.
#  // Also, displays the random state obtained from a get_randstate call.
#  // In subsequent simulations, set_randstate can be called with the same
#  // value to reproduce the same sequence of transactions.
#
#  function new(string name, uvm_component parent);
#
#    super.new(name, parent);
#
#    blocking_put_port=new("blocking_put_port", this);
#    
#    uvm_report_info("uvm_stimulus", {"rand state is ", get_randstate()});
#
#  endfunction
#
#
#  local bit m_stop;
#
#
#  // Function: generate_stimulus
#  //
#  // Generate up to max_count transactions of type T.
#  // If t is not specified, a default instance of T is allocated and used.
#  // If t is specified, that transaction is used when randomizing. It must
#  // be a subclass of T.
#  //
#  // max_count is the maximum number of transactions to be
#  // generated. A value of zero indicates no maximum - in
#  // this case, generate_stimulus will go on indefinitely
#  // unless stopped by some other process
#  //
#  // The transactions are cloned before they are sent out 
#  // over the blocking_put_port
#
#  virtual task generate_stimulus(T t=null, int max_count=0);
#
#    T temp;
#    
#    if (t == null)
#      t = new;
#    
#    for (int i=0; (max_count == 0 || i < max_count) && !m_stop; i++) begin
#
#       if (! t.randomize() )
#          uvm_report_warning ("RANDFL", "Randomization failed in generate_stimulus");
#      
#       $cast(temp, t.clone());
#       uvm_report_info("stimulus generation", temp.convert2string()); 
#       blocking_put_port.put(temp);
#    end
#  endtask
#  
#
#  // Function: stop_stimulus_generation
#  //
#  // Stops the generation of stimulus.
#  // If a subclass of this method has forked additional
#  // processes, those processes will also need to be
#  // stopped in an overridden version of this method
#  
#  virtual function void stop_stimulus_generation;
#    m_stop = 1;
#  endfunction
#  
#
#  virtual function string get_type_name();
#    return type_name;
#  endfunction
#
#endclass : uvm_random_stimulus



#//
#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
"""
Title: UVM Run-Time Phases

The run-time schedule is the pre-defined phase schedule
which runs concurrently to the `UVMRunPhase` global run phase.
By default, all `UVMComponent`s using the run-time schedule
are synchronized with respect to the pre-defined phases in the schedule.
It is possible for components to belong to different domains
in which case their schedules can be unsynchronized.

The names of the UVM phases (which will be returned by get_name() for a
phase instance) match the class names specified below with the "uvm_"
and "_phase" removed.  For example, the main phase corresponds to the
UVMMainPhase class below and has the name "main", which means that
the following can be used to call foo() at the start of main phase::

    def phase_started(self, phase):
        if phase.get_name()=="main":
            self.foo()

The run-time phases are executed in the sequence they are specified below.
"""

from .uvm_task_phase import UVMTaskPhase
from .uvm_debug import uvm_debug


class UVMPreResetPhase(UVMTaskPhase):
    """
    Class: UVMPreResetPhase

    Before reset is asserted.

    `UVMTaskPhase` that calls the
    `UVMComponent.pre_reset_phase` method. This phase starts at the
    same time as the `UVMRunPhase` unless a user defined phase is inserted
    in front of this phase.

    Upon Entry:
    - Indicates that power has been applied but not necessarily valid or stable.
    - There should not have been any active clock edges
      before entry into this phase.

    Typical Uses:
    - Wait for power good.
    - Components connected to virtual interfaces should initialize
      their output to X's or Z's.
    - Initialize the clock signals to a valid value
    - Assign reset signals to X (power-on reset).
    - Wait for reset signal to be asserted
      if not driven by the verification environment.

    Exit Criteria:
    - Reset signal, if driven by the verification environment,
      is ready to be asserted.
    - Reset signal, if not driven by the verification environment, is asserted.
    """
    async def exec_task(self,comp, phase):
        await comp.pre_reset_phase(phase)

    m_inst = None
    type_name = "uvm_pre_reset_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if cls.m_inst is None:
            cls.m_inst = UVMPreResetPhase()
            return cls.m_inst

    def __init__(self, name="pre_reset"):
        super().__init__(name)

    def get_type_name(self):
        return UVMPreResetPhase.type_name


class UVMResetPhase(UVMTaskPhase):
    """
    Class: UVMResetPhase

    Reset is asserted.

    `UVMTaskPhase` that calls the
    `UVMComponent.reset_phase` method.

    Upon Entry:
    - Indicates that the hardware reset signal is ready to be asserted.

    Typical Uses:
    - Assert reset signals.
    - Components connected to virtual interfaces should drive their output
      to their specified reset or idle value.
    - Components and environments should initialize their state variables.
    - Clock generators start generating active edges.
    - De-assert the reset signal(s)  just before exit.
    - Wait for the reset signal(s) to be de-asserted.

    Exit Criteria:
    - Reset signal has just been de-asserted.
    - Main or base clock is working and stable.
    - At least one active clock edge has occurred.
    - Output signals and state variables have been initialized.
    """

    async def exec_task(self,comp, phase):
        await comp.reset_phase(phase)

    m_inst = None
    type_name = "uvm_reset_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if cls.m_inst is None:
            cls.m_inst = UVMResetPhase()
            return cls.m_inst

    def __init__(self, name="reset"):
        super().__init__(name)

    def get_type_name(self):
        return UVMResetPhase.type_name

class UVMPostResetPhase(UVMTaskPhase):
    """
    Class: uvm_post_reset_phase

    After reset is de-asserted.

    `UVMTaskPhase` that calls the
    `UVMComponent.post_reset_phase` method.

    Upon Entry:
    - Indicates that the DUT reset signal has been de-asserted.

    Typical Uses:
    - Components should start behavior appropriate for reset being inactive.
      For example, components may start to transmit idle transactions
      or interface training and rate negotiation.
      This behavior typically continues beyond the end of this phase.

    Exit Criteria:
    - The testbench and the DUT are in a known, active state.
    """

    async def exec_task(self,comp, phase):
        await comp.post_reset_phase(phase)

    m_inst = None
    type_name = "uvm_post_reset_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if(cls.m_inst is None):
            cls.m_inst = UVMPostResetPhase()
            return cls.m_inst

    def __init__(self, name="post_reset"):
        super().__init__(name)

    def get_type_name(self):
        return UVMPostResetPhase.type_name


class UVMPreConfigurePhase(UVMTaskPhase):
    """
    Class: UVMPreConfigurePhase

    Before the DUT is configured by the SW.

    `UVMTaskPhase` that calls the
    `UVMComponent.pre_configure_phase` method.

    Upon Entry:
    - Indicates that the DUT has been completed reset
     and is ready to be configured.

    Typical Uses:
    - Procedurally modify the DUT configuration information as described
      in the environment (and that will be eventually uploaded into the DUT).
    - Wait for components required for DUT configuration to complete
      training and rate negotiation.

    Exit Criteria:
    - DUT configuration information is defined.
    """

    async def exec_task(self,comp, phase):
        await comp.pre_configure_phase(phase)

    m_inst = None
    type_name = "uvm_pre_configure_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if(cls.m_inst is None):
            cls.m_inst = UVMPreConfigurePhase()
            return cls.m_inst

    def __init__(self, name="pre_configure"):
        super().__init__(name)

    def get_type_name(self):
        return UVMPreConfigurePhase.type_name


class UVMConfigurePhase(UVMTaskPhase):
    """
    Class: UVMConfigurePhase

    The SW configures the DUT.

    `UVMTaskPhase` that calls the
    `UVMComponent.configure_phase` method.

    Upon Entry:
    - Indicates that the DUT is ready to be configured.

    Typical Uses:
    - Components required for DUT configuration execute transactions normally.
    - Set signals and program the DUT and memories
      (e.g. read/write operations and sequences)
      to match the desired configuration for the test and environment.

    Exit Criteria:
    - The DUT has been configured and is ready to operate normally.
    """

    async def exec_task(self,comp, phase):
        await comp.configure_phase(phase)

    m_inst = None
    type_name = "uvm_configure_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if(cls.m_inst is None):
            cls.m_inst = UVMConfigurePhase()
            return cls.m_inst

    def __init__(self, name="configure"):
        super().__init__(name)

    def get_type_name(self):
        return UVMConfigurePhase.type_name


class UVMPostConfigurePhase(UVMTaskPhase):
    """
    Class: UVMPostConfigurePhase

    After the SW has configured the DUT.

    `UVMTaskPhase` that calls the
    `UVMComponent.post_configure_phase` method.

    Upon Entry:
    - Indicates that the configuration information has been fully uploaded.

    Typical Uses:
    - Wait for configuration information to fully propagate and take effect.
    - Wait for components to complete training and rate negotiation.
    - Enable the DUT.
    - Sample DUT configuration coverage.

    Exit Criteria:
    - The DUT has been fully configured and enabled
      and is ready to start operating normally.
    """

    async def exec_task(self,comp, phase):
        await comp.post_configure_phase(phase)

    m_inst = None
    type_name = "uvm_post_configure_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if(cls.m_inst is None):
            cls.m_inst = UVMPostConfigurePhase()
            return cls.m_inst

    def __init__(self, name="post_configure"):
        super().__init__(name)

    def get_type_name(self):
        return UVMPostConfigurePhase.type_name


class UVMPreMainPhase(UVMTaskPhase):
    """
    Class: UVMPreMainPhase

    Before the primary test stimulus starts.

    `UVMTaskPhase` that calls the
    `UVMComponent.pre_main_phase` method.

    Upon Entry:
    - Indicates that the DUT has been fully configured.

    Typical Uses:
    - Wait for components to complete training and rate negotiation.

    Exit Criteria:
    - All components have completed training and rate negotiation.
    - All components are ready to generate and/or observe normal stimulus.
    """

    async def exec_task(self,comp, phase):
        await comp.pre_main_phase(phase)

    m_inst = None
    type_name = "uvm_pre_main_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if(cls.m_inst is None):
            cls.m_inst = UVMPreMainPhase()
            return cls.m_inst

    def __init__(self, name="pre_main"):
        super().__init__(name)

    def get_type_name(self):
        return UVMPreMainPhase.type_name


class UVMMainPhase(UVMTaskPhase):
    """
    Class: UVMMainPhase

    Primary test stimulus.

    `UVMTaskPhase` that calls the
    `UVMComponent.main_phase` method.

    Upon Entry:
    - The stimulus associated with the test objectives is ready to be applied.

    Typical Uses:
    - Components execute transactions normally.
    - Data stimulus sequences are started.
    - Wait for a time-out or certain amount of time,
      or completion of stimulus sequences.

    Exit Criteria:
    - Enough stimulus has been applied to meet the primary
      stimulus objective of the test.
    """

    async def exec_task(self,comp, phase):
        uvm_debug(self, 'exec_task', 'yield main_phase for ' + comp.get_name())
        await comp.main_phase(phase)

    m_inst = None  # local static UVMMainPhase
    type_name = "UVMMainPhase"

    @classmethod
    def get(cls):
        """
        Function: get
        Returns the singleton phase handle
        Returns:
        """
        if cls.m_inst is None:
            cls.m_inst = UVMMainPhase()
        return cls.m_inst

    def __init__(self, name="main"):
        super().__init__(name)

    def get_type_name(self):
        return UVMMainPhase.type_name



class UVMPostMainPhase(UVMTaskPhase):
    """
    Class: UVMPostMainPhase

    After enough of the primary test stimulus.

    `UVMTaskPhase` that calls the
    `UVMComponent.post_main_phase` method.

    Upon Entry:
    - The primary stimulus objective of the test has been met.

    Typical Uses:
    - Included for symmetry.

    Exit Criteria:
    - None.
    """

    async def exec_task(self,comp, phase):
        await comp.post_main_phase(phase)

    m_inst = None
    type_name = "uvm_post_main_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if(cls.m_inst is None):
            cls.m_inst = UVMPostMainPhase()
            return cls.m_inst

    def __init__(self, name="post_main"):
        super().__init__(name)

    def get_type_name(self):
        return UVMPostMainPhase.type_name


class UVMPreShutdownPhase(UVMTaskPhase):
    """
    Class: UVMPreShutdownPhase

    Before things settle down.

    `UVMTaskPhase` that calls the
    `UVMComponent.pre_shutdown_phase` method.

    Upon Entry:
    - None.

    Typical Uses:
    - Included for symmetry.

    Exit Criteria:
    - None.
    """

    async def exec_task(self,comp, phase):
        await comp.pre_shutdown_phase(phase)

    m_inst = None
    type_name = "uvm_pre_shutdown_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if(cls.m_inst is None):
            cls.m_inst = UVMPreShutdownPhase()
            return cls.m_inst

    def __init__(self, name="pre_shutdown"):
        super().__init__(name)

    def get_type_name(self):
        return UVMPreShutdownPhase.type_name


class UVMShutdownPhase(UVMTaskPhase):
    """
    Class: uvm_shutdown_phase

    Letting things settle down.

    `UVMTaskPhase` that calls the
    `UVMComponent.shutdown_phase` method.

    Upon Entry:
    - None.

    Typical Uses:
    - Wait for all data to be drained out of the DUT.
    - Extract data still buffered in the DUT,
      usually through read/write operations or sequences.

    Exit Criteria:
    - All data has been drained or extracted from the DUT.
    - All interfaces are idle.
    """

    async def exec_task(self,comp, phase):
        await comp.shutdown_phase(phase)

    m_inst = None
    type_name = "uvm_shutdown_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if(cls.m_inst is None):
            cls.m_inst = UVMShutdownPhase()
            return cls.m_inst

    def __init__(self, name="shutdown"):
        super().__init__(name)

    def get_type_name(self):
        return UVMShutdownPhase.type_name


class UVMPostShutdownPhase(UVMTaskPhase):
    """
    Class: UVMPostShutdownPhase

    After things have settled down.

    `UVMTaskPhase` that calls the
    `UVMComponent.post_shutdown_phase` method.  The end of this phase is
    synchronized to the end of the `UVMRunPhase` phase unless a user defined
    phase is added after this phase.

    Upon Entry:
    - No more "data" stimulus is applied to the DUT.

    Typical Uses:
    - Perform final checks that require run-time access to the DUT
      (e.g. read accounting registers or dump the content of memories).

    Exit Criteria:
    - All run-time checks have been satisfied.
    - The `UVMRunPhase` phase is ready to end.
    """

    async def exec_task(self,comp, phase):
        await comp.post_shutdown_phase(phase)

    m_inst = None
    type_name = "uvm_post_shutdown_phase"

    @classmethod
    def get(cls):
        """
           Function: get
           Returns the singleton phase handle
        Returns:
        """
        if cls.m_inst is None:
            cls.m_inst = UVMPostShutdownPhase()
            return cls.m_inst

    def __init__(self, name="post_shutdown"):
        super().__init__(name)

    def get_type_name(self):
        return UVMPostShutdownPhase.type_name



#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from ..base.uvm_globals import (UVM_ERROR, UVM_FATAL, UVM_INFO, UVM_NONE, UVM_WARNING,
                                uvm_report_enabled, uvm_report_error, uvm_report_fatal,
                                uvm_report_info, uvm_report_warning)

import inspect
from inspect import getframeinfo, stack



UVM_REPORT_DISABLE_FILE = False
UVM_REPORT_DISABLE_LINE = False

UVM_LINE_WIDTH = 120
UVM_NUM_LINES = 120


def uvm_file():
    return "<TODO uvm_file>"


if UVM_REPORT_DISABLE_FILE:
    def uvm_file():
        return ""


def uvm_line():
    return "<TODO uvm_line>"


if UVM_REPORT_DISABLE_LINE:
    def uvm_line():
        return 0

#//------------------------------------------------------------------------------
#//
#// Title: Report Macros
#//
#// This set of macros provides wrappers around the uvm_report_* <Reporting>
#// functions. The macros serve two essential purposes:
#//
#// - To reduce the processing overhead associated with filtered out messages,
#//   a check is made against the report's verbosity setting and the action
#//   for the id/severity pair before any string formatting is performed. This
#//   affects only `uvm_info reports.
#//
#// - The `__FILE__ and `__LINE__ information is automatically provided to the
#//   underlying uvm_report_* call. Having the file and line number from where
#//   a report was issued aides in debug. You can disable display of file and
#//   line information in reports by defining UVM_REPORT_DISABLE_FILE_LINE on
#//   the command line.
#//
#// The macros also enforce a verbosity setting of UVM_NONE for warnings, errors
#// and fatals so that they cannot be mistakenly turned off by setting the
#// verbosity level too low (warning and errors can still be turned off by
#// setting the actions appropriately).
#//
#// To use the macros, replace the previous call to uvm_report_* with the
#// corresponding macro.
#//
#//| //Previous calls to uvm_report_*
#//| uvm_report_info("MYINFO1", $sformatf("val: %0d", val), UVM_LOW);
#//| uvm_report_warning("MYWARN1", "This is a warning");
#//| uvm_report_error("MYERR", "This is an error");
#//| uvm_report_fatal("MYFATAL", "A fatal error has occurred");
#//
#// The above code is replaced by
#//
#//| //New calls to `uvm_*
#//| `uvm_info("MYINFO1", $sformatf("val: %0d", val), UVM_LOW)
#//| `uvm_warning("MYWARN1", "This is a warning")
#//| `uvm_error("MYERR", "This is an error")
#//| `uvm_fatal("MYFATAL", "A fatal error has occurred")
#//
#// Macros represent text substitutions, not statements, so they should not be
#// terminated with semi-colons.


#//----------------------------------------------------------------------------
#// Group:  Basic Messaging Macros
#//----------------------------------------------------------------------------


#// MACRO: `uvm_info
#//
#// Calls uvm_report_info if ~VERBOSITY~ is lower than the configured verbosity of
#// the associated reporter. ~ID~ is given as the message tag and ~MSG~ is given as
#// the message text. The file and line are also sent to the uvm_report_info call.
#//
#// |`uvm_info(ID, MSG, VERBOSITY)

def uvm_info(ID, MSG, VERBOSITY):
    if uvm_report_enabled(VERBOSITY, UVM_INFO, ID):
        caller = getframeinfo(stack()[1][0])
        fname = caller.filename
        lineno = caller.lineno
        f_locals = inspect.currentframe().f_back.f_locals  # type: ignore
        if 'self' in f_locals:
            parent_self = f_locals['self']
            if hasattr(parent_self, 'uvm_report_info'):
                parent_self.uvm_report_info(ID, MSG, VERBOSITY, fname, lineno, "", 1)
            else:
                uvm_report_info(ID, MSG, VERBOSITY, fname, lineno, "", 1)
        else:
            uvm_report_info(ID, MSG, VERBOSITY, fname, lineno, "", 1)


#// MACRO: `uvm_warning
#//
#// Calls uvm_report_warning with a verbosity of UVM_NONE. The message cannot
#// be turned off using the reporter's verbosity setting, but can be turned off
#// by setting the action for the message.  ~ID~ is given as the message tag and
#// ~MSG~ is given as the message text. The file and line are also sent to the
#// uvm_report_warning call.
#//
#// |`uvm_warning(ID, MSG)

def uvm_warning(ID, MSG):
    if uvm_report_enabled(UVM_NONE, UVM_WARNING,ID):
        caller = getframeinfo(stack()[1][0])
        fname = caller.filename
        lineno = caller.lineno
        f_locals = inspect.currentframe().f_back.f_locals  # type: ignore
        if 'self' in f_locals:
            parent_self = f_locals['self']
            if hasattr(parent_self, 'uvm_report_warning'):
                parent_self.uvm_report_warning(ID, MSG, UVM_NONE, fname, lineno, "", 1)
            else:
                uvm_report_warning(ID, MSG, UVM_NONE, fname, lineno, "", 1)
        else:
            uvm_report_warning(ID, MSG, UVM_NONE, fname, lineno, "", 1)

#// MACRO: `uvm_error
#//
#// Calls uvm_report_error with a verbosity of UVM_NONE. The message cannot
#// be turned off using the reporter's verbosity setting, but can be turned off
#// by setting the action for the message.  ~ID~ is given as the message tag and
#// ~MSG~ is given as the message text. The file and line are also sent to the
#// uvm_report_error call.
#//
#// |`uvm_error(ID, MSG)

def uvm_error(ID, MSG):
    if uvm_report_enabled(UVM_NONE, UVM_ERROR,ID):
        caller = getframeinfo(stack()[1][0])
        fname = caller.filename
        lineno = caller.lineno
        f_locals = inspect.currentframe().f_back.f_locals  # type: ignore
        if 'self' in f_locals:
            parent_self = f_locals['self']
            if hasattr(parent_self, 'uvm_report_error'):
                parent_self.uvm_report_error(ID, MSG, UVM_NONE, fname, lineno, "", 1)
            else:
                uvm_report_error(ID, MSG, UVM_NONE, fname, lineno, "", 1)
        else:
            uvm_report_error(ID, MSG, UVM_NONE, fname, lineno, "", 1)

#// MACRO: `uvm_fatal
#//
#// Calls uvm_report_fatal with a verbosity of UVM_NONE. The message cannot
#// be turned off using the reporter's verbosity setting, but can be turned off
#// by setting the action for the message.  ~ID~ is given as the message tag and
#// ~MSG~ is given as the message text. The file and line are also sent to the
#// uvm_report_fatal call.
#//
#// |`uvm_fatal(ID, MSG)


def uvm_fatal(ID, MSG):
    if uvm_report_enabled(UVM_NONE, UVM_FATAL,ID):
        caller = getframeinfo(stack()[1][0])
        fname = caller.filename
        lineno = caller.lineno
        f_locals = inspect.currentframe().f_back.f_locals  # type: ignore
        if 'self' in f_locals:
            parent_self = f_locals['self']
            if hasattr(parent_self, 'uvm_report_fatal'):
                parent_self.uvm_report_fatal(ID, MSG, UVM_NONE, fname, lineno, "", 1)
            else:
                uvm_report_fatal(ID, MSG, UVM_NONE, fname, lineno, "", 1)
        else:
            uvm_report_fatal(ID, MSG, UVM_NONE, fname, lineno, "", 1)

#// MACRO: `uvm_info_context
#//
#//| `uvm_info_context(ID, MSG, VERBOSITY, RO)
#//
#// Operates identically to `uvm_info but requires that the
#// context, or <uvm_report_object>, in which the message is printed be
#// explicitly supplied as a macro argument.

def uvm_info_context(ID, MSG, VERBOSITY, RO):
    if RO.uvm_report_enabled(VERBOSITY, UVM_INFO, ID):
        caller = getframeinfo(stack()[1][0])
        fname = caller.filename
        lineno = caller.lineno
        RO.uvm_report_info(ID, MSG, VERBOSITY, fname, lineno, "", 1)


#// MACRO: `uvm_warning_context
#//
#//| `uvm_warning_context(ID, MSG, RO)
#//
#// Operates identically to `uvm_warning but requires that the
#// context, or <uvm_report_object>, in which the message is printed be
#// explicitly supplied as a macro argument.
#
#`define uvm_warning_context(ID, MSG, RO) \
#   begin \
#     if (RO.uvm_report_enabled(UVM_NONE,UVM_WARNING,ID)) \
#       RO.uvm_report_warning (ID, MSG, UVM_NONE, uvm_file(), uvm_line(), "", 1); \
#   end


#// MACRO: `uvm_error_context
#//
#//| `uvm_error_context(ID, MSG, RO)
#//
#// Operates identically to `uvm_error but requires that the
#// context, or <uvm_report_object> in which the message is printed be
#// explicitly supplied as a macro argument.
#
#`define uvm_error_context(ID, MSG, RO) \
#   begin \
#     if (RO.uvm_report_enabled(UVM_NONE,UVM_ERROR,ID)) \
#       RO.uvm_report_error (ID, MSG, UVM_NONE, uvm_file(), uvm_line(), "", 1); \
#   end
def uvm_error_context(ID, MSG, RO):
    if RO.uvm_report_enabled(UVM_NONE, UVM_ERROR, ID):
        caller = getframeinfo(stack()[1][0])
        fname = caller.filename
        lineno = caller.lineno
        RO.uvm_report_error(ID, MSG, UVM_NONE, fname, lineno, "", 1)


#// MACRO: `uvm_fatal_context
#//
#//| `uvm_fatal_context(ID, MSG, RO)
#//
#// Operates identically to `uvm_fatal but requires that the
#// context, or <uvm_report_object>, in which the message is printed be
#// explicitly supplied as a macro argument.
#
#`define uvm_fatal_context(ID, MSG, RO) \
#   begin \
#     if (RO.uvm_report_enabled(UVM_NONE,UVM_FATAL,ID)) \
#       RO.uvm_report_fatal (ID, MSG, UVM_NONE, uvm_file(), uvm_line(), "", 1); \
#   end


#//----------------------------------------------------------------------------
#// Group:  Message Trace Macros
#//----------------------------------------------------------------------------

#// MACRO- `uvm_message_begin
#//
#// Undocumented. Library internal use.
#//

#`define uvm_message_begin(SEVERITY, ID, MSG, VERBOSITY, FILE, LINE, RM) \
#   begin \
#     if (uvm_report_enabled(VERBOSITY,SEVERITY,ID)) begin \
#       uvm_report_message __uvm_msg; \
#       if (RM == null) RM = uvm_report_message::new_report_message(); \
#       __uvm_msg = RM; \
#       __uvm_msg.set_report_message(SEVERITY, ID, MSG, VERBOSITY, FILE, LINE, "");


#// MACRO- `uvm_message_end
#//
#// Undocumented. Library internal use.
#//
#//

#`define uvm_message_end \
#       uvm_process_report_message(__uvm_msg); \
#     end \
#   end

#// MACRO- `uvm_message_context_begin
#//
#// Undocumented. Library internal use.
#//

#`define uvm_message_context_begin(SEVERITY, ID, MSG, VERBOSITY, FILE, LINE, RO, RM) \
#   begin \
#     uvm_report_object __report_object; \
#     __report_object = RO; \
#     if (__report_object.uvm_report_enabled(VERBOSITY,SEVERITY,ID)) begin \
#       uvm_report_message __uvm_msg; \
#       if (RM == null) RM = uvm_report_message::new_report_message(); \
#       __uvm_msg = RM; \
#       __uvm_msg.set_report_message(SEVERITY, ID, MSG, VERBOSITY, FILE, LINE, "");
#
#
#// MACRO- `uvm_message_context_end
#//
#// Undocumented. Library internal use.
#//
#//

#`define uvm_message_context_end \
#       __report_object.uvm_process_report_message(__uvm_msg); \
#     end \
#   end


#// MACRO: `uvm_info_begin
#//
#// |`uvm_info_begin(ID, MSG, VERBOSITY, RM = __uvm_msg)
#//
#
#`define uvm_info_begin(ID, MSG, VERBOSITY, RM = __uvm_msg) \
#   `uvm_message_begin(UVM_INFO, ID, MSG, VERBOSITY, uvm_file(), uvm_line(), RM)
#
#// MACRO: `uvm_info_end
#//
#// This macro pair provides the ability to add elements to messages.
#//
#// |`uvm_info_end
#//
#// Example usage is shown here.
#//
#// |...
#// |task my_task();
#// |   ...
#// |   `uvm_info_begin("MY_ID", "This is my message...", UVM_LOW)
#// |     `uvm_message_add_tag("my_color", "red")
#// |     `uvm_message_add_int(my_int, UVM_DEC)
#// |     `uvm_message_add_string(my_string)
#// |     `uvm_message_add_object(my_obj)
#// |   `uvm_info_end
#// |   ...
#// |endtask
#//
#
#`define uvm_info_end \
#   `uvm_message_end
#
#
#// MACRO: `uvm_warning_begin
#//
#// |`uvm_warning_begin(ID, MSG, RM = __uvm_msg)
#//
#
#`define uvm_warning_begin(ID, MSG, RM = __uvm_msg) \
#   `uvm_message_begin(UVM_WARNING, ID, MSG, UVM_NONE, uvm_file(), uvm_line(), RM)
#
#// MACRO: `uvm_warning_end
#//
#// This macro pair operates identically to <`uvm_info_begin>/<`uvm_info_end> with
#// exception that the message severity is <UVM_WARNING> and has no verbosity threshold.
#//
#// |`uvm_warning_end
#//
#// The usage shown in <`uvm_info_end> works identically for this pair.
#//
#
#`define uvm_warning_end \
#   `uvm_message_end
#
#
#// MACRO: `uvm_error_begin
#//
#// |`uvm_error_begin(ID, MSG, RM = __uvm_msg)
#//
#
#`define uvm_error_begin(ID, MSG, RM = __uvm_msg) \
#   `uvm_message_begin(UVM_ERROR, ID, MSG, UVM_NONE, uvm_file(), uvm_line(), RM)
#
#
#// MACRO: `uvm_error_end
#//
#// This macro pair operates identically to <`uvm_info_begin>/<`uvm_info_end> with
#// exception that the message severity is <UVM_ERROR> and has no verbosity threshold.
#//
#// |`uvm_error_end
#//
#// The usage shown in <`uvm_info_end> works identically for this pair.
#//
#
#`define uvm_error_end \
#   `uvm_message_end
#
#
#// MACRO: `uvm_fatal_begin
#//
#// |`uvm_fatal_begin(ID, MSG, RM = __uvm_msg)
#//
#
#`define uvm_fatal_begin(ID, MSG, RM = __uvm_msg) \
#   `uvm_message_begin(UVM_FATAL, ID, MSG, UVM_NONE, uvm_file(), uvm_line(), RM)
#
#
#// MACRO: `uvm_fatal_end
#//
#// This macro pair operates identically to <`uvm_info_begin>/<`uvm_info_end> with
#// exception that the message severity is <UVM_FATAL> and has no verbosity threshold.
#//
#// |`uvm_fatal_end
#//
#// The usage shown in <`uvm_info_end> works identically for this pair.
#//
#
#`define uvm_fatal_end \
#   `uvm_message_end
#
#
#// MACRO: `uvm_info_context_begin
#//
#// |`uvm_info_context_begin(ID, MSG, UVM_NONE, RO, RM = __uvm_msg)
#//
#
#`define uvm_info_context_begin(ID, MSG, VERBOSITY, RO, RM = __uvm_msg) \
#   `uvm_message_context_begin(UVM_INFO, ID, MSG, VERBOSITY, uvm_file(), uvm_line(), RO, RM)
#
#
#// MACRO: `uvm_info_context_end
#//
#// |`uvm_info_context_end
#//
#// This macro pair operates identically to <`uvm_info_begin>/<`uvm_info_end>, but
#// requires that the context, or <uvm_report_object> in which the message is printed
#// be explicitly supplied as a macro argument.
#//
#
#`define uvm_info_context_end \
#   `uvm_message_context_end
#
#
#// MACRO: `uvm_warning_context_begin
#//
#// |`uvm_warning_context_begin(ID, MSG, RO, RM = __uvm_msg)
#//
#
#`define uvm_warning_context_begin(ID, MSG, RO, RM = __uvm_msg) \
#   `uvm_message_context_begin(UVM_WARNING, ID, MSG, UVM_NONE, uvm_file(), uvm_line(), RO, RM)
#
#// MACRO: `uvm_warning_context_end
#//
#// |`uvm_warning_context_end
#//
#// This macro pair operates identically to <`uvm_warning_begin>/<`uvm_warning_end>, but
#// requires that the context, or <uvm_report_object> in which the message is printed
#// be explicitly supplied as a macro argument.
#//
#
#`define uvm_warning_context_end \
#   `uvm_message_context_end
#
#
#// MACRO: `uvm_error_context_begin
#//
#// |`uvm_error_context_begin(ID, MSG, RO, RM = __uvm_msg)
#//
#
#`define uvm_error_context_begin(ID, MSG, RO, RM = __uvm_msg) \
#   `uvm_message_context_begin(UVM_ERROR, ID, MSG, UVM_NONE, uvm_file(), uvm_line(), RO, RM)
#
#
#// MACRO: `uvm_error_context_end
#//
#// |`uvm_error_context_end
#//
#// This macro pair operates identically to <`uvm_error_begin>/<`uvm_error_end>, but
#// requires that the context, or <uvm_report_object> in which the message is printed
#// be explicitly supplied as a macro argument.
#//
#
#`define uvm_error_context_end \
#   `uvm_message_context_end
#
#
#// MACRO: `uvm_fatal_context_begin
#//
#// |`uvm_fatal_context_begin(ID, MSG, RO, RM = __uvm_msg)
#//
#
#`define uvm_fatal_context_begin(ID, MSG, RO, RM = __uvm_msg) \
#   `uvm_message_context_begin(UVM_FATAL, ID, MSG, UVM_NONE, uvm_file(), uvm_line(), RO, RM)
#
#
#// MACRO: `uvm_fatal_context_end
#//
#// |`uvm_fatal_context_end
#//
#// This macro pair operates identically to <`uvm_fatal_begin>/<`uvm_fatal_end>, but
#// requires that the context, or <uvm_report_object> in which the message is printed
#// be explicitly supplied as a macro argument.
#//
#
#`define uvm_fatal_context_end \
#   `uvm_message_context_end
#
#
#//----------------------------------------------------------------------------
#// Group:  Message Element Macros
#//----------------------------------------------------------------------------
#
#
#// MACRO: `uvm_message_add_tag
#//
#// |`uvm_message_add_tag(NAME, VALUE, ACTION=(UVM_LOG|UVM_RM_RECORD))
#//
#
#`define uvm_message_add_tag(NAME, VALUE, ACTION=(UVM_LOG|UVM_RM_RECORD)) \
#    __uvm_msg.add_string(NAME, VALUE, ACTION);
#
#
#// MACRO: `uvm_message_add_int
#//
#// |`uvm_message_add_int(VAR, RADIX, LABEL = "", ACTION=(UVM_LOG|UVM_RM_RECORD))
#//
#
#`define uvm_message_add_int(VAR, RADIX, LABEL="", ACTION=(UVM_LOG|UVM_RM_RECORD)) \
#    if (LABEL == "") \
#      __uvm_msg.add_int(`"VAR`", VAR, $bits(VAR), RADIX, ACTION); \
#    else \
#      __uvm_msg.add_int(LABEL, VAR, $bits(VAR), RADIX, ACTION);
#
#
#// MACRO: `uvm_message_add_string
#//
#// |`uvm_message_add_string(VAR, LABEL = "", ACTION=(UVM_LOG|UVM_RM_RECORD))
#//
#
#`define uvm_message_add_string(VAR, LABEL="", ACTION=(UVM_LOG|UVM_RM_RECORD)) \
#    if (LABEL == "") \
#      __uvm_msg.add_string(`"VAR`", VAR, ACTION); \
#    else \
#      __uvm_msg.add_string(LABEL, VAR, ACTION);
#
#
#// MACRO: `uvm_message_add_object
#//
#// These macros allow the user to provide elements that are associated with
#// <uvm_report_message>s.  Separate macros are provided such that the
#// user can supply arbitrary string/string pairs using <`uvm_message_add_tag>,
#// integral types along with a radix using <`uvm_message_add_int>, string
#// using <`uvm_message_add_string> and <uvm_object>s using
#// <`uvm_message_add_object>.
#//
#// |`uvm_message_add_object(VAR, LABEL = "", ACTION=(UVM_LOG|UVM_RM_RECORD))
#//
#// Example usage is shown in <`uvm_info_end>.
#//
#
#`define uvm_message_add_object(VAR, LABEL="", ACTION=(UVM_LOG|UVM_RM_RECORD)) \
#    if (LABEL == "") \
#      __uvm_msg.add_object(`"VAR`", VAR, ACTION); \
#    else \
#      __uvm_msg.add_object(LABEL, VAR, ACTION);
#



#//----------------------------------------------------------------------
#// Copyright 2010-2011 Mentor Graphics Corporation
#// Copyright 2010 Synopsys, Inc.
#// Copyright 2010-2018 Cadence Design Systems, Inc.
#// Copyright 2014-2015 NVIDIA Corporation
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

#//----------------------------------------------------------------------
#// Title -- NODOCS -- Interface Masks
#//
#// Each of the following macros is a mask that identifies which
#// interfaces a particular port requires or export provides.  The
#// interfaces are identified by bit position and can be OR'ed together
#// for combination ports/exports.  The mask is used to do run-time
#// interface type checking of port/export connections.
#//----------------------------------------------------------------------

#// MACRO -- NODOCS -- `UVM_TLM_NB_FW_MASK
#//
#// Define Non blocking Forward mask onehot assignment = 'b001
UVM_TLM_NB_FW_MASK = (1<<0)
#
#// MACRO: `UVM_TLM_NB_BW_MASK
#//
#// Define Non blocking backward mask onehot assignment = 'b010
UVM_TLM_NB_BW_MASK = (1<<1)
#
#// MACRO: `UVM_TLM_B_MASK
#//
#// Define blocking mask onehot assignment = 'b100
UVM_TLM_B_MASK = (1<<2)



#//----------------------------------------------------------------------
#// Copyright 2010-2011 Mentor Graphics Corporation
#// Copyright 2010 Synopsys, Inc.
#// Copyright 2010-2018 Cadence Design Systems, Inc.
#// Copyright 2015 NVIDIA Corporation
#// Copyright 2014 Cisco Systems, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

#//----------------------------------------------------------------------
#// Title -- NODOCS -- TLM2 ports
#//
#// The following defines TLM2 port classes.
#//
#//----------------------------------------------------------------------

#//
#// Class providing the blocking transport port.
#// The port can be bound to one export.
#// There is no backward path for the blocking transport.
from uvm.base.uvm_port_base import UVMPortBase
from uvm.tlm1.uvm_tlm_imps import UVM_PORT_COMMON
from uvm.tlm2.uvm_tlm2_defines import UVM_TLM_B_MASK, UVM_TLM_NB_FW_MASK,\
    UVM_TLM_NB_BW_MASK
from uvm.tlm2.uvm_tlm2_imps import UVM_TLM_B_TRANSPORT_IMP,\
    UVM_TLM_NB_TRANSPORT_FW_IMP, UVM_TLM_NB_TRANSPORT_BW_IMP

class UVMTLMBTransportPort(UVMPortBase):
    pass

UVMTLMBTransportPort = UVM_PORT_COMMON(UVMTLMBTransportPort,  # type: ignore
        UVM_TLM_B_MASK, "uvm_tlm_b_transport_port")
UVM_TLM_B_TRANSPORT_IMP('m_if', UVMTLMBTransportPort)

#
#// class -- NODOCS -- uvm_tlm_nb_transport_fw_port
#//
#// Class providing the non-blocking backward transport port.
#// Transactions received from the producer, on the forward path, are
#// sent back to the producer on the backward path using this
#// non-blocking transport port.
#// The port can be bound to one export.
#//
class UVMTLMNbTransportFwPort(UVMPortBase):
    pass

UVMTLMNbTransportFwPort = UVM_PORT_COMMON(UVMTLMNbTransportFwPort,  # type: ignore
        UVM_TLM_NB_FW_MASK, "uvm_tlm_nb_transport_fw_port")
UVM_TLM_NB_TRANSPORT_FW_IMP('m_if', UVMTLMNbTransportFwPort)

#// class: uvm_tlm_nb_transport_bw_port
#//
#// Class providing the non-blocking backward transport port.
#// Transactions received from the producer, on the forward path, are
#// sent back to the producer on the backward path using this
#// non-blocking transport port
#// The port can be bound to one export.
#//

class UVMTLMNbTransportBwPort(UVMPortBase):
    pass

UVMTLMNbTransportBwPort = UVM_PORT_COMMON(UVMTLMNbTransportBwPort,  # type: ignore
        UVM_TLM_NB_BW_MASK, "uvm_tlm_nb_transport_bw_port")
UVM_TLM_NB_TRANSPORT_BW_IMP('m_if', UVMTLMNbTransportBwPort)



#//----------------------------------------------------------------------
#// Copyright 2010-2011 Mentor Graphics Corporation
#// Copyright 2010 Synopsys, Inc.
#// Copyright 2010-2018 Cadence Design Systems, Inc.
#// Copyright 2015-2018 NVIDIA Corporation
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------


#from uvm.base.uvm_tlm_b_initiator_socket_base import *
#from uvm.base.uvm_tlm_b_target_socket_base import *
#from uvm.base.uvm_tlm_nb_initiator_socket_base import *
#from uvm.base.uvm_tlm_nb_target_socket_base import *
#from uvm.base.uvm_tlm_b_passthrough_initiator_socket_base import *
#from uvm.base.uvm_tlm_b_passthrough_target_socket_base import *
#from uvm.base.uvm_tlm_nb_passthrough_initiator_socket_base import *
#from uvm.base.uvm_tlm_nb_passthrough_target_socket_base import *

from ..macros.uvm_message_defines import (uvm_error_context, uvm_error,
    uvm_fatal)
from .uvm_tlm2_sockets_base import (UVMTLMBInitiatorSocketBase,
    UVMTLMBTargetSocketBase)
from .uvm_tlm2_imps import (UVM_TLM_B_TRANSPORT_IMP)
from ..base.sv import sv

"""
TLM2 Sockets

Each uvm_tlm_*_socket class is derived from a corresponding
uvm_tlm_*_socket_base class.  The base class contains most of the
implementation of the class, The derived classes (in this file)
contain the connection semantics.

Sockets come in several flavors: Each socket is either an initiator or a
target, a pass-through or a terminator. Further, any particular socket
implements either the blocking interfaces or the nonblocking interfaces.
Terminator sockets are used on initiators and targets as well as
interconnect components as shown in the figure above. Pass-through
 sockets are used to enable connections to cross hierarchical boundaries.

There are eight socket types: the cross of blocking and nonblocking,
pass-through and termination, target and initiator

Sockets are specified based on what they are (IS-A)
and what they contains (HAS-A).
IS-A and HAS-A are types of object relationships.
IS-A refers to the inheritance relationship and
 HAS-A refers to the ownership relationship.
For example if you say D is a B that means that D is derived from base B.
If you say object A HAS-A B that means that B is a member of A.
"""


from uvm.tlm2.uvm_tlm2_sockets_base import UVMTLMBPassthroughInitiatorSocketBase,\
    UVMTLMBPassthroughTargetSocketBase, UVMTLMBTargetSocketBase,\
    UVMTLMBInitiatorSocketBase, UVMTLMNbInitiatorSocketBase,\
    UVMTLMNbPassthroughInitiatorSocketBase, UVMTLMNbPassthroughTargetSocketBase,\
    UVMTLMNbTargetSocketBase
from uvm.base.sv import cat
from uvm.macros.uvm_message_defines import uvm_error
from uvm.tlm2.uvm_tlm2_imps import UVM_TLM_NB_TRANSPORT_FW_IMP,\
    UVM_TLM_B_TRANSPORT_IMP, UVMTLMNbTransportBwImp
from uvm.tlm2.uvm_tlm2_ports import UVMTLMNbTransportBwPort


#//----------------------------------------------------------------------
#// Class -- NODOCS -- UVMTLMBInitiatorSocket
#//
#// IS-A forward port; has no backward path except via the payload
#// contents
#//----------------------------------------------------------------------
class UVMTLMBInitiatorSocket(UVMTLMBInitiatorSocketBase):

    #  // Function: new
    #  // Construct a new instance of this socket
    def __init__(self, name, parent):
        super().__init__(name, parent)

    #// Function: Connect
    #//
    #// Connect this socket to the specified <uvm_tlm_b_target_socket>
    def connect(self, provider):
        valid_providers = (UVMTLMBPassthroughInitiatorSocketBase,
            UVMTLMBPassthroughTargetSocketBase, UVMTLMBTargetSocketBase)

        super().connect(provider)

        if not isinstance(provider, valid_providers):
            c = self.get_comp()
            uvm_error_context(self.get_type_name(),
                "type mismatch in connect -- connection cannot be completed", c)

#//----------------------------------------------------------------------
#// Class -- NODOCS -- uvm_tlm_b_target_socket
#//
#// IS-A forward imp; has no backward path except via the payload
#// contents.
#//
#// The component instantiating this socket must implement
#// a b_transport() method with the following signature
#//
#//|   task b_transport(T t, uvm_tlm_time delay);
#//
#//----------------------------------------------------------------------
class UVMTLMBTargetSocket(UVMTLMBTargetSocketBase):

    #// Function: new
    #// Construct a new instance of this socket
    #// ~imp~ is a reference to the class implementing the
    #// b_transport() method.
    #// If not specified, it is assume to be the same as ~parent~.
    def __init__(self, name, parent, imp = None):
        super().__init__(name, parent)
        self.m_imp = parent if imp == None else imp
        if (self.m_imp == None):
            uvm_error("UVM/TLM2/NOIMP", cat("b_target socket ", name,
                                     " has no implementation"));

    #// Function: Connect
    #//
    #// Connect this socket to the specified <uvm_tlm_b_initiator_socket>
    def connect(self, provider):
        super().connect(provider)

        c = self.get_comp();
        # TODO:
#         uvm_error_context(get_type_name(),
#             "You cannot call connect() on a target termination socket", c)

UVM_TLM_B_TRANSPORT_IMP('m_imp', UVMTLMBTargetSocket)

#//----------------------------------------------------------------------
#// Class -- NODOCS -- uvm_tlm_nb_initiator_socket
#//
#// IS-A forward port; HAS-A backward imp
#//
#// The component instantiating this socket must implement
#// a nb_transport_bw() method with the following signature
#//
#//|   function uvm_tlm_sync_e nb_transport_bw(T t, ref P p, input uvm_tlm_time delay);
#//
#//----------------------------------------------------------------------

class UVMTLMNbInitiatorSocket(UVMTLMNbInitiatorSocketBase):

    #// Function: new
    #// Construct a new instance of this socket
    #// ~imp~ is a reference to the class implementing the
    #// nb_transport_bw() method.
    #// If not specified, it is assume to be the same as ~parent~.
    def __init__(self, name, parent, imp=None):
        super().__init__(name, parent)

        if (imp is None):
            imp = parent
        else:
            uvm_error("UVM/TLM2/NOIMP", cat("nb_initiator socket ", name,
                                     " has no implementation"));
        self.bw_imp = UVMTLMNbTransportBwImp("bw_imp", imp)

    #// Function: Connect
    #//
    #// Connect this socket to the specified <uvm_tlm_nb_target_socket>
    def connect(self, provider):
        valid_providers = (UVMTLMNbPassthroughInitiatorSocketBase,
            UVMTLMNbPassthroughTargetSocketBase,
            UVMTLMNbTargetSocketBase)

        super().connect(provider)

        if not isinstance(provider, valid_providers):
            msg = 'Req type: ' + str(valid_providers) + ' Got: ' + str(provider)
            uvm_error_context(self.get_type_name(),
                "type mismatch in connect -- connection cannot be done: " + msg,
                self.get_comp())
            return

        sock = []
        if sv.cast(sock, provider, UVMTLMNbPassthroughInitiatorSocketBase):
            sock[0].bw_export.connect(self.bw_imp)
            return

        if sv.cast(sock, provider, UVMTLMNbPassthroughTargetSocketBase):
            sock[0].bw_port.connect(self.bw_imp)
            return

        if sv.cast(sock, provider, UVMTLMNbTargetSocketBase):
            sock[0].bw_port.connect(self.bw_imp)
            return
        uvm_fatal("UVMTLMNbInitiatorSocket",
            "UVM Internal error. Should not go here")
        

#//----------------------------------------------------------------------
#// Class -- NODOCS -- uvm_tlm_nb_target_socket
#//
#// IS-A forward imp; HAS-A backward port
#//
#// The component instantiating this socket must implement
#// a nb_transport_fw() method with the following signature
#//
#//|   function uvm_tlm_sync_e nb_transport_fw(T t, ref P p, input uvm_tlm_time delay);
#//
#//----------------------------------------------------------------------

class UVMTLMNbTargetSocket(UVMTLMNbTargetSocketBase):

    #// Function: new
    #// Construct a new instance of this socket
    #// ~imp~ is a reference to the class implementing the
    #// nb_transport_fw() method.
    #// If not specified, it is assume to be the same as ~parent~.
    def __init__(self, name, parent, imp=None):
        super().__init__(name, parent)
        self.m_imp = parent if imp is None else imp
        self.bw_port = UVMTLMNbTransportBwPort("bw_port", self.get_comp())

        if self.m_imp is None:
            uvm_error("UVM/TLM2/NOIMP", cat("nb_target socket ", name,
                                     " has no implementation"))

    #// Function: connect
    #//
    #// Connect this socket to the specified <uvm_tlm_nb_initiator_socket>
    def connect(self, provider):
        super().connect(provider)
        # TODO:
        #uvm_error_context(self.get_type_name(),
        #    "You cannot call connect() on a target termination socket",
        #    self.get_comp())

UVM_TLM_NB_TRANSPORT_FW_IMP('m_imp', UVMTLMNbTargetSocket)


#//----------------------------------------------------------------------
#// Class -- NODOCS -- uvm_tlm_b_passthrough_initiator_socket
#//
#// IS-A forward port;
#//----------------------------------------------------------------------
#
class UVMTLMBPassthroughInitiatorSocket(UVMTLMBPassthroughInitiatorSocketBase):

    def __init__(self, name, parent):
        super().__init__(name, parent)

    #// Function : connect
    #//
    #// Connect this socket to the specified <uvm_tlm_b_target_socket>
    def connect(self, provider):
        valid_providers = (UVMTLMBPassthroughInitiatorSocketBase,
            UVMTLMBPassthroughTargetSocketBase, UVMTLMBTargetSocketBase)

        super().connect(provider)

        if not isinstance(provider, valid_providers):
            # TODO:
            # uvm_error_context(get_type_name(),
            #    "type mismatch in connect -- connection cannot be completed",
            #    self.get_comp())
            pass

#//----------------------------------------------------------------------
#// Class: uvm_tlm_b_passthrough_target_socket
#//
#// IS-A forward export;
#//----------------------------------------------------------------------
class UVMTLMBPassthroughTargetSocket(UVMTLMBPassthroughTargetSocketBase):

    def __init__(self, name, parent):
        super().__init__(name, parent)

    #// Function : connect
    #//
    #// Connect this socket to the specified <uvm_tlm_b_initiator_socket>
    def connect(self, provider):
        valid_providers = (UVMTLMBPassthroughTargetSocketBase,
            UVMTLMBTargetSocketBase)

        super().connect(provider)

        if not isinstance(provider, valid_providers):
            # TODO:
            # uvm_error_context(self.get_type_name(),
            #    "type mismatch in connect -- connection cannot be completed",
            #    self.get_comp())
            pass


#//----------------------------------------------------------------------
#// Class -- NODOCS -- uvm_tlm_nb_passthrough_initiator_socket
#//
#// IS-A forward port; HAS-A backward export
#//----------------------------------------------------------------------

class UVMTLMNbPassthroughInitiatorSocket(UVMTLMNbPassthroughInitiatorSocketBase):

    def __init__(self, name, parent):
        super().__init__(name, parent)


    #// Function : connect
    #//
    #// Connect this socket to the specified <uvm_tlm_nb_target_socket>
    def connect(self, provider):
        valid_providers = (UVMTLMNbPassthroughInitiatorSocketBase,
            UVMTLMNbPassthroughTargetSocketBase, UVMTLMNbTargetSocketBase)

        super().connect(provider)

        if not isinstance(provider, valid_providers):
            # TODO:
            # uvm_error_context(self.get_type_name(),
            #     "type mismatch in connect -- connection cannot be completed",
            #     self.get_comp())
            pass

#//----------------------------------------------------------------------
#// Class -- NODOCS -- uvm_tlm_nb_passthrough_target_socket
#//
#// IS-A forward export; HAS-A backward port
#//----------------------------------------------------------------------

class UVMTLMNbPassthroughTargetSocket(UVMTLMNbPassthroughTargetSocketBase):

    def __init__(self, name, parent):
        super().__init__(name, parent)

    #// Function: connect
    #//
    #// Connect this socket to the specified <uvm_tlm_nb_initiator_socket>
    def connect(self, provider):
        valid_providers = (UVMTLMNbPassthroughTargetSocketBase,
            UVMTLMNbTargetSocketBase)

        super().connect(provider)

        if not isinstance(provider, valid_providers):
            # TODO:
            # uvm_error_context(self.get_type_name(),
            #     "type mismatch in connect -- connection cannot be completed",
            #     self.get_comp())
            pass



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer
from uvm.base import UVMComponent, UVMPhase, uvm_top
#from uvm.base.uvm_root import uvm_top
from uvm.tlm1.uvm_tlm_fifos import UVMTLMFIFO
from uvm.macros import *
from producer import producer
from consumer import consumer

class top(UVMComponent):
    #  producer #(packet) p1
    #  producer #(packet) p2
    #  uvm_tlm_fifo #(packet) f
    #  consumer #(packet) c

    def __init__(self, name, parent=None):
        UVMComponent.__init__(self, name, parent)
        UVMPhase.m_phase_trace = True
        self.p1 = producer("producer1", self)
        self.p2 = producer("producer2", self)
        self.f = UVMTLMFIFO("fifo", self)
        self.c = consumer("consumer", self)

        # Create connections between components
        self.p1.out.connect(self.c.input)
        self.p2.out.connect(self.f.blocking_put_export)
        self.c.out.connect(self.f.get_export)
        self.error = False

    def end_of_elaboration_phase(self, phase):
        uvm_top.print_topology()

    
    async def run_phase(self, phase):
        print("top run_Phase started XYZ")
        phase.raise_objection(self)
        await Timer(1000, "NS")
        phase.drop_objection(self)

    def check_phase(self, phase):
        if self.c.count == 0:
            self.error = True
            self.uvm_report_error("COUNT_ERR", "Count was 0")

uvm_component_utils(top)

class top2(UVMComponent):

    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.prods = []
        self.prods2 = []
        self.cons = []
        self.fifos = []
        self.n = 128

        for i in range(self.n):
            self.prods.append(producer("producer" + str(i), self))
            self.prods2.append(producer("producer2_" + str(i), self))
            self.cons.append(consumer("consumer" + str(i), self))
            self.fifos.append(UVMTLMFIFO("fifo" + str(i), self))
        for i in range(self.n):
            self.prods[i].out.connect(self.cons[i].input)
            self.prods2[i].out.connect(self.fifos[i].blocking_put_export)
            self.cons[i].out.connect(self.fifos[i].get_export)

    async def run_phase(self, phase):
        print("top2 run_Phase started XYZ")
        phase.raise_objection(self)
        await Timer(1000, "NS")
        phase.drop_objection(self)

    def check_phase(self, phase):
        for i in range(self.n):
            if self.cons[i].count == 0:
                self.error = True
                self.uvm_report_error("COUNT_ERR", "Count was 0")

uvm_component_utils(top2)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm import *

class apb_rw(UVMSequenceItem):

    READ = 0
    WRITE = 1
    #   typedef enum {READ, WRITE} kind_e
    #   rand bit   [31:0] addr
    #   rand logic [31:0] data
    #   rand kind_e kind;


    def __init__(self, name = "apb_rw"):
        super().__init__(name)
        self.kind = apb_rw.READ
        self.addr = 0
        self.data = 0


    def convert2string(self):
        return sv.sformatf("kind=%s addr=%0h data=%0h",self.kind,self.addr,self.data)


uvm_object_utils_begin(apb_rw)
uvm_field_int('addr', UVM_ALL_ON | UVM_NOPACK)
uvm_field_int('data', UVM_ALL_ON | UVM_NOPACK)
#uvm_field_enum(kind_e,kind, UVM_ALL_ON | UVM_NOPACK)
uvm_object_utils_end(apb_rw)



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2014 NVIDIA Corportation
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

from cocotb.triggers import Event

from .sv import sv, wait
from .uvm_object import UVMObject
from ..macros.uvm_message_defines import uvm_warning
from .uvm_globals import uvm_zero_delay
from .uvm_queue import UVMQueue



class UVMEventBase(UVMObject):
    """
    The UVMEventBase class is an abstract wrapper class around the
    `cocotb.triggers.Event` class.  It provides some additional services
    such as setting self.callbacks and maintaining the number of waiters.
    """

    type_name = "uvm_event_base"

    def __init__(self, name=""):
        """
        Creates a new event object.

        Args:
            name (str): Name of this event.
        """
        super().__init__(name)
        self.m_event = Event()
        self.num_waiters = 0
        self.on = False
        self.on_event = Event("on_event_" + name)
        self.trigger_time = 0
        self.callbacks = UVMQueue()  # list[uvm_event_callback]
        self.m_waiters = 0
        self.m_value_changed_event = Event("value_changed_event_" + name)


    def set_value(self, key, value):
        """
        Sets a value inside this event object and triggers value_changed_event.

        Args:
            key (str): Member variable name
            value (any): Value to set for that member variable.
        """
        setattr(self, key, value)
        self.m_value_changed_event.set()

    def set(self):
        if self.m_waiters > 0:
            self.on_event.set()


    async def wait(self):
        if self.m_waiters == 0:
            self.on_event.clear()

        self.m_waiters += 1
        await self.on_event.wait()
        self.m_waiters -= 1
        if self.m_waiters == 0:
            self.on_event.clear()


    async def wait_on(self, delta=False):
        """
        Waits for the event to be activated for the first time.

        If the event has already been triggered, this task returns immediately.
        If `delta` is set, the caller will be forced to wait a single delta #0
        before returning. This prevents the caller from returning before
        previously waiting processes have had a chance to resume.

        Once an event has been triggered, it will be remain "on" until the event
        is `UVMEvent.reset`.

        Args:
            delta (bool): If True, wait for one zero delay
        """
        if self.on is True:
            if delta is True:
                #0
                await uvm_zero_delay()
            await uvm_zero_delay()  # TODO tpoikela: redundant, should be removed?
            return
        self.num_waiters += 1
        await wait(lambda: self.on is True, self.m_value_changed_event)


    async def wait_off(self, delta=False):
        """
        If the event has already triggered and is "on", this task waits for the
        event to be turned "off" via a call to `reset`.

        If the event has not already been triggered, this task returns immediately.
        If `delta` is set, the caller will be forced to wait a single delta #0
        before returning. This prevents the caller from returning before
        previously waiting processes have had a chance to resume.

        Args:
            delta (bool): If True, wait for one zero delay
        """
        if self.on is False:
            if delta is True:
                await uvm_zero_delay()
            return
        self.num_waiters += 1
        await wait(lambda: self.on is False, self.m_value_changed_event)


    async def wait_trigger(self):
        """
        Waits for the event to be triggered.

        If one process calls wait_trigger in the same delta as another process
        calls `UVMEvent.trigger`, a race condition occurs. If the call to wait occurs
        before the trigger, this method will return in this delta. If the wait
        occurs after the trigger, this method will not return until the next
        trigger, which may never occur and thus cause deadlock.

        """
        self.num_waiters += 1
        await self.m_event.wait()
        self.m_event.clear()


    async def wait_ptrigger(self):
        """
        Waits for a persistent trigger of the event. Unlike `UVMEvent.wait_trigger`, this
        views the trigger as persistent within a given time-slice and thus avoids
        certain race conditions. If this method is called after the trigger but
        within the same time-slice, the caller returns immediately.
        """
        if self.m_event.fired:
            return
        self.num_waiters += 1
        await self.m_event.wait()
        self.m_event.clear()

    def get_trigger_time(self):
        """
        Gets the time that this event was last triggered. If the event has not bee
        triggered, or the event has been reset, then the trigger time will be 0.

        Returns:
            int: Time when this event was triggered.
        """
        return self.trigger_time


    def is_on(self):
        """
        Indicates whether the event has been triggered since it was last reset.

        A return of 1 indicates that the event has triggered.

        Returns:
        """
        return self.on

    def is_off(self):
        """
        Indicates whether the event has been triggered or been reset.

        A return of 1 indicates that the event has not been triggered.

        Returns:
        """
        return not self.on

    def reset(self, wakeup=False):
        """
        Resets the event to its off state. If `wakeup` is set, then all processes
        currently waiting for the event are activated before the reset.

        No self.callbacks are called during a reset.

        Args:
            wakeup (bool):
        """
        if wakeup:
            self.m_event.set()
        self.m_event = Event()
        self.num_waiters = 0
        self.set_value("on", False)
        self.trigger_time = 0


    def cancel(self):
        """
        Decrements the number of waiters on the event.

        This is used if a process that is waiting on an event is disabled or
        activated by some other means.
        """
        if self.num_waiters > 0:
            self.num_waiters -= 1


    def get_num_waiters(self):
        """
        Returns the number of processes waiting on the event.
        """
        return self.num_waiters


    def get_type_name(self) -> str:
        return UVMEventBase.type_name

    #    virtual function void do_print (uvm_printer printer)
    #        printer.print_field_int("self.num_waiters",
    #            self.num_waiters, $bits(self.num_waiters), UVM_DEC, ".", "int")
    #        printer.print_field_int("on", self.on, $bits(self.on), UVM_BIN, ".", "bit")
    #        printer.print_time("self.trigger_time", self.trigger_time)
    #        printer.m_scope.down("self.callbacks")
    #        foreach(self.callbacks[e]) begin
    #            printer.print_object($sformatf("[%0d]",e), self.callbacks[e], "[")
    #        end
    #        printer.m_scope.up()
    #    endfunction


    #    virtual function void do_copy (uvm_object rhs)
    #        uvm_event_base e
    #        super.do_copy(rhs)
    #        if(!$cast(e, rhs) || (e is None)) return
    #
    #        self.m_event = e.self.m_event
    #        self.num_waiters = e.self.num_waiters
    #        self.set_value("on", e.on)
    #        self.trigger_time = e.self.trigger_time
    #        self.callbacks.delete()
    #        self.callbacks = e.self.callbacks
    #
    #    endfunction

    #endclass



class UVMEvent(UVMEventBase):  # (type T=uvm_object) extends uvm_event_base
    """
    CLASS: UVMEvent

    The UVMEvent class is an extension of the abstract `UVMEventBase` class.

    The optional parameter ~T~ allows the user to define a data type which
    can be passed during an event trigger.
    """
    type_name = "uvm_event"

    def __init__(self, name="", T=None):
        """
        Creates a new event object.

        Args:
            name (str): Name of the event.
            T (any): Optional data type.
        """
        super().__init__(name)
        self.trigger_data = None
        self.T = T


    async def wait_trigger_data(self):  # output T data)
        """
        Task: wait_trigger_data

        This method calls <uvm_event_base::wait_trigger> followed by `get_trigger_data`.
        Returns:
            any: Trigger data.
        """
        await self.wait_trigger()
        return self.get_trigger_data()


    async def wait_ptrigger_data(self, data):
        """
        This method calls <uvm_event_base::wait_ptrigger> followed by `get_trigger_data`.

        Args:
            data:
        Returns:
            any: Data used to trigger this event.
        """
        await self.wait_ptrigger()
        trig_data = self.get_trigger_data()
        data.append(trig_data)
        return trig_data

    #    //------------//
    #    // triggering //
    #    //------------//

    def trigger(self, data=None):
        """
        Triggers the event, resuming all waiting processes.

        An optional `data` argument can be supplied with the enable to provide
        trigger-specific information.

        Args:
            data (any): Data associated with the trigger.
        """
        skip = False
        if self.callbacks.size() > 0:
            for i in range(0, self.callbacks.size()):
                #uvm_event_callback#(T) tmp=self.callbacks[i]
                tmp = self.callbacks.get(i)
            skip = skip + tmp.pre_trigger(self, data)
        if skip is False:
            self.m_event.set(data)
            if self.callbacks.size() > 0:
                for i in range(0, self.callbacks.size()):
                    #uvm_event_callback#(T) tmp=self.callbacks[i]
                    tmp = self.callbacks[i]
                    tmp.post_trigger(self,data)  # type: ignore
            self.num_waiters = 0
            self.set_value("on", True)
            self.trigger_time = sv.realtime()
            self.trigger_data = data


    def get_trigger_data(self):
        """
        Gets the data, if any, provided by the last call to `trigger`.

        Returns:
            any: Data used to trigger this event.
        """
        return self.trigger_data

    def get_type_name(self):
        return UVMEvent.type_name


    def add_callback(self, cb, append=True):
        """
        Registers a callback object, `cb`, with this event. The callback object
        may include pre_trigger and post_trigger functionality. If `append` is set
        to 1, the default, `cb` is added to the back of the callback list. Otherwise,
        `cb` is placed at the front of the callback list.

        Args:
            cb:
            append (bool): If True, append to the end. Otherwise insert to front.
        """
        if cb in self.callbacks:
            uvm_warning("CBRGED","add_callback: Callback already registered. Ignoring.")
            return

        if append is True:
            self.callbacks.push_back(cb)
        else:
            self.callbacks.insert(0, cb)


    def delete_callback(self, cb):
        """
        Unregisters the given callback, ~cb~, from this event.

        Args:
            cb (UVMCallback): Callback to delete from this event.
        """
        for i in range(len(self.callbacks)):
            if cb == self.callbacks[i]:
                self.callbacks.delete(i)
                return
        uvm_warning("CBNTFD", "delete_callback: Callback not found. Ignoring delete request.")

    def do_print(self, printer):
        super().do_print(printer)
        printer.print_object("trigger_data", self.trigger_data)


    #    virtual function void do_copy (uvm_object rhs)
    #        uvm_event#(T) e
    #        super.do_copy(rhs)
    #        if(!$cast(e, rhs) || (e is None)) return
    #        self.trigger_data = e.trigger_data
    #    endfunction // do_copy

    #   virtual function uvm_object create(string name="")
    #        uvm_event#(T) v
    #        v=new(name)
    #        return v
    #    endfunction



#//
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import Timer

from uvm.base.uvm_component import UVMComponent
from uvm.macros import uvm_fatal, uvm_component_utils
from uvm.base.uvm_config_db import UVMConfigDb

#//
#// Generic phase timer
#//
#// All time-out values are interprted in ns
#//
#// To set time-out values:
#//
#// - For one phase:
#//
#//   uvm_config_db#(time)::set(null, "global_timer.main", "timeout", 100);
#//
#// - For multiple phases:
#//
#//   uvm_config_db#(time)::set(null, "global_timer.pre*", "timeout", 100);
#//


class tb_timer(UVMComponent):

    m_global = None

    def __init__(self, name, parent=None):
        super().__init__(name, parent)


    
    async def run_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "run", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in run phase")

    #
    
    async def pre_reset_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "pre_reset", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in pre_reset phase")

    #
    
    async def reset_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "reset", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in reset phase")

    #
    
    async def post_reset_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "post_reset", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in post_reset phase")

    #
    
    async def pre_configure_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "pre_configure", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in pre_configure phase")

    #
    
    async def configure_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "configure", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in configure phase")

    #
    
    async def post_configure_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "post_configure", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in post_configure phase")

    #
    
    async def pre_main_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "pre_main", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in pre_main phase")

    #
    
    async def main_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "main", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in main phase")

    #
    
    async def post_main_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "post_main", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in post_main phase")

    #
    
    async def pre_shutdown_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "pre_shutdown", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in pre_shutdown phase")

    #
    
    async def shutdown_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "shutdown", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in shutdown phase")

    #
    
    async def post_shutdown_phase(self, phase):
        t = []
        if (UVMConfigDb.get(self, "post_shutdown", "timeout", t) and t[0] > 0):
            await Timer(t[0], "NS")
            uvm_fatal("TIMEOUT", "Time-out expired in post_shutdown phase")



uvm_component_utils(tb_timer)

#tb_timer.m_global = tb_timer("global_timer", None)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

#//----------------------------------------------------------------------
#// This example illustrates basic hierarchy construction and test phasing
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer
from uvm import (UVMComponent, sv, UVMEnv, run_test, uvm_info, uvm_fatal,
    UVM_NONE, uvm_error, UVMCoreService, UVM_ERROR)

CORRECT_PHASES = ['build', 'connect', 'end_of_elaboration',
    'start_of_simulation',
    'run', 'pre_reset', 'reset', 'post_reset', 'pre_configure', 'configure',
    'post_configure', 'pre_main', 'main', 'post_main', 'pre_shutdown',
    'shutdown', 'post_shutdown',
    'extract', 'check', 'report', 'final']

#  //Create a topology
#  //            top
#  //       |            |
#  //     u1(A)         u2(A)
#  //    |     |      |        |
#  //   b1(B) d1(D)  b1(B)    d1(D)

# Define new base class with phase_started/ended callbacks.

test_duration = 10


class VerifComp(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.phases_started = []
        self.phases_ended = []
        self.correct = False

    def phase_started(self, phase):
        self.phases_started.append(phase.get_name())

    def phase_ended(self, phase):
        self.phases_ended.append(phase.get_name())

    def final_phase(self, phase):
        if self.phases_started != CORRECT_PHASES:
            uvm_error("PHASES_STARTED_FAIL", self.get_name() + ": " +
                    str(self.phases_started))
        else:
            self.correct = True
            uvm_info("PHASES_STARTED_OK", self.get_name() + ": " + str(self.phases_started),
                UVM_NONE)


# No run phase
class D(VerifComp):
    def __init__(self, name, parent):
        super().__init__(name, parent)

    def build_phase(self, phase):
        sv.display("%0t: %0s:  build", sv.time("NS"), self.get_full_name())

    def end_of_elaboration_phase(self, phase):
        sv.display("%0t: %0s:  end_of_elaboration", sv.time("NS"), self.get_full_name())

    def start_of_simulation_phase(self, phase):
        sv.display("%0t: %0s:  start_of_simulation", sv.time("NS"), self.get_full_name())

    def extract_phase(self, phase):
        sv.display("%0t: %0s:  extract", sv.time("NS"), self.get_full_name())

    def check_phase(self, phase):
        sv.display("%0t: %0s:  check", sv.time("NS"), self.get_full_name())

    def report_phase(self, phase):
        sv.display("%0t: %0s:  report", sv.time("NS"), self.get_full_name())

    #  endclass


#  //Has run phase
class B(VerifComp):
    #    rand logic [7:0] delay
    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.delay = 0
        self.rand('delay', range(0, (1 << 8) - 1))

    def build_phase(self, phase):
        sv.display("%0t: %0s:  build", sv.time("NS"), self.get_full_name())

    def end_of_elaboration_phase(self, phase):
        sv.display("%0t: %0s:  end_of_elaboration", sv.time("NS"), self.get_full_name())

    def start_of_simulation_phase(self, phase):
        sv.display("%0t: %0s:  start_of_simulation", sv.time("NS"), self.get_full_name())

    def extract_phase(self, phase):
        sv.display("%0t: %0s:  extract", sv.time("NS"), self.get_full_name())

    def check_phase(self, phase):
        sv.display("%0t: %0s:  check", sv.time("NS"), self.get_full_name())

    def report_phase(self, phase):
        sv.display("%0t: %0s:  report", sv.time("NS"), self.get_full_name())

    
    async def run_phase(self, phase):
        sv.display("%0t: %0s:  start run phase", sv.time("NS"), self.get_full_name())
        await Timer(self.delay, "NS")
        sv.display("%0t: %0s:  end run phase", sv.time("NS"), self.get_full_name())


# Has run phase and contains subcomponents
class A(VerifComp):
    #    rand B b1
    #    rand D d1
    #    rand logic [7:0] delay
    def __init__(self, name, parent):
        super().__init__(name,parent)
        self.b1 = B("b1", self)
        self.d1 = D("d1", self)
        self.delay = 0
        self.rand('delay', range(0, (1 << 8) - 1))

    def build_phase(self, phase):
        sv.display("%0t: %0s:  build", sv.time("NS"), self.get_full_name())

    def end_of_elaboration_phase(self, phase):
        sv.display("%0t: %0s:  end_of_elaboration", sv.time("NS"), self.get_full_name())

    def start_of_simulation_phase(self, phase):
        sv.display("%0t: %0s:  start_of_simulation", sv.time("NS"), self.get_full_name())

    def extract_phase(self, phase):
        sv.display("%0t: %0s:  extract", sv.time("NS"), self.get_full_name())

    def check_phase(self, phase):
        sv.display("%0t: %0s:  check", sv.time("NS"), self.get_full_name())

    def report_phase(self, phase):
        sv.display("%0t: %0s:  report", sv.time("NS"), self.get_full_name())

    
    async def run_phase(self, phase):
        sv.display("%0t: %0s:  start run phase", sv.time("NS"), self.get_full_name())
        await Timer(self.delay, "NS")
        sv.display("%0t: %0s:  end run phase", sv.time("NS"), self.get_full_name())


class AA(VerifComp):
    #    rand A a
    def __init__(self, name, parent):
        super().__init__(name,parent)
        self.a = A("a", self)


# Top level contains two A components
class top(UVMEnv):
    #    rand AA a1
    #    rand AA a2

    def __init__(self, name, parent):
        super().__init__(name,parent)
        self.a1 = AA("a1", self)
        self.a2 = AA("a2", self)

    def build_phase(self, phase):
        sv.display("%0t: %0s:  build", sv.time("NS"), self.get_full_name())

    def end_of_elaboration_phase(self, phase):
        sv.display("%0t: %0s:  end_of_elaboration", sv.time("NS"), self.get_full_name())

    def start_of_simulation_phase(self, phase):
        sv.display("%0t: %0s:  start_of_simulation", sv.time("NS"), self.get_full_name())

    def extract_phase(self, phase):
        sv.display("%0t: %0s:  extract", sv.time("NS"), self.get_full_name())

    def check_phase(self, phase):
        sv.display("%0t: %0s:  check", sv.time("NS"), self.get_full_name())

    def report_phase(self, phase):
        sv.display("%0t: %0s:  report", sv.time("NS"), self.get_full_name())

    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        sv.display("%0t: %0s:  start run phase", sv.time("NS"), self.get_full_name())
        await Timer(test_duration, "NS")
        sv.display("%0t: %0s:  end run phase", sv.time("NS"), self.get_full_name())
        phase.drop_objection(self)



@cocotb.test()
async def initial_begin(dut):
    t = top("top", None)
    # Randomize all of the delays
    t.randomize()  # cast to 'void' removed
    await run_test()
    if sv.realtime("NS") != test_duration:
        uvm_fatal("TIME_ERR", "Exp: {}, Got: {}".format(
            test_duration, sv.realtime("NS")))
    svr = UVMCoreService.get().get_report_server()
    num_errors = svr.get_severity_count(UVM_ERROR)
    if num_errors > 0:
        uvm_fatal("SERVER_HAD_ERRORS", "Num UVM_ERRORS: {}".format(num_errors))



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from uvm import (UVMConfigDb, run_test, UVMObject)
from my_env_pkg import my_env

#//Begindoc: Automated Configuration
#//This test will focus on testing auto configurations using the uvm_component related config methods.


#//Walk through the test:
#//Configuration settings are stored in each component instance. 

#//Then a search is made to verify that those components full names match the field_name that had been set.


@cocotb.test()
async def module_top(dut):
    obj = UVMObject("my_conf_obj")

    # set configuration prior to creating the environment
    UVMConfigDb.set(None, "topenv.*.u1", "v", 30)
    UVMConfigDb.set(None, "topenv.inst2.u1", "v", 10)
    UVMConfigDb.set(None, "*", "recording_detail", 0)
    UVMConfigDb.set(None, "*", "myaa[foo]", "hi")
    UVMConfigDb.set(None, "*", "myaa[bar]", "bye")
    UVMConfigDb.set(None, "*", "myaa[foobar]", "howdy")
    UVMConfigDb.set(None, "topenv.inst1.u1", "myaa[foo]", "boo")
    UVMConfigDb.set(None, "topenv.inst1.u1", "myaa[foobar]", "boobah")
    UVMConfigDb.set(None, "topenv.inst1.u1", "my_conf_obj", obj)
    UVMConfigDb.set(None, "topenv.inst1.u1", "tag", 'I am tagged')
    topenv = my_env("topenv", None)
    await run_test()
